//
// Created by philipp on 15.01.18.
//

#ifndef NETWORKINGLIB_DATAGRAMRECEIVER_H
#define NETWORKINGLIB_DATAGRAMRECEIVER_H

#include "Networking.h"
#include "Stream.h"
#include "Socket.h"
#include "Message.h"

namespace networking
{
namespace message
{

template<typename Message>
class DatagramReceiver : public std::enable_shared_from_this<DatagramReceiver<Message>>
{
private:
    struct PrivateTag
    {
    };

public:
    using Ptr = std::shared_ptr<DatagramReceiver<Message>>;

    using Udp = boost::asio::ip::udp;
    using Endpoint = Udp::endpoint;
    using Socket = Udp::socket;

    using ReceiveHandler = std::function<
        void(const error::ErrorCode & error,
             Message & message,
             const std::string & host,
             std::uint16_t port)>;

    static Ptr create(Networking & net, std::uint16_t bindingPort, std::size_t maxMessageSize = 512)
    {
        return std::make_shared<DatagramReceiver<Message>>(PrivateTag{}, net, bindingPort, maxMessageSize);
    }

    DatagramReceiver(PrivateTag, Networking & net, std::uint16_t bindingPort, std::size_t maxMessageSize)
        : net(net)
          , bindingPort(bindingPort)
          , socket(net.getIoService(), Udp::v4())
          , buffer(maxMessageSize)
    {}

    bool isReceiving() const noexcept
    { return receiving; };

    Message receive(std::string & host, std::uint16_t & port, const time::Duration & timeout)
    {
        if (receiving)
            throw error::Busy{};

        utils::RAIIObject receivingFlagReset{[this]
                                             { receiving = false; }};
        receiving = true;
        newSocket();
        return message::receiveDatagramFrom<Message>(net, socket, buffer, host, port, timeout);
    }

    void asyncReceive(const time::Duration & timeout, const ReceiveHandler & handler)
    {
        if (receiving)
            throw error::Busy{};

        auto self = this->shared_from_this();
        auto state = std::make_shared<AsyncState>(self, handler);

        receiving = true;
        newSocket();

        message::asyncReceiveDatagramFrom<Message>(
            net, socket, buffer, timeout,
            [state](const error::ErrorCode& error,
                    protocol::PlatoonMessage & message,
                    const std::string & senderHost,
                    std::uint16_t senderPort)
            {
                state->self->receiving = false;
                state->handler(error, message, senderHost, senderPort);
            });
    }

    void stop()
    {
        if (!receiving)
            return;

        closeable::Closer<Socket>::close(socket);
    }

private:
    Networking & net;
    std::uint16_t bindingPort;
    Socket socket;
    std::vector<char> buffer;
    std::atomic<bool> receiving{false};

    void newSocket()
    {
        socket = Socket(net.getIoService(), Udp::v4());
        socket.set_option(boost::asio::socket_base::reuse_address{true});
        socket.bind(Endpoint(Udp::v4(), bindingPort));
    }

    struct AsyncState
    {
        AsyncState(Ptr self, const ReceiveHandler & handler)
            : self(self)
              , handler(handler)
              , receivingFlagReset([self]
                                   { self->receiving = false; })
        {}

        Ptr self;
        ReceiveHandler handler;
        utils::RAIIObject receivingFlagReset;
    };
};

}
}

#endif //NETWORKINGLIB_DATAGRAMRECEIVER_H
