//
// Created by philipp on 01.12.17.
//

#ifndef PROTOCOL_MESSAGES_H
#define PROTOCOL_MESSAGES_H

#include <cstdint>
#include <vector>
#include <boost/system/error_code.hpp>
#include <boost/asio/write.hpp>
#include <boost/asio/read.hpp>
#include "Stream.h"
#include "Networking.h"
#include "Socket.h"
#include <boost/algorithm/string/replace.hpp>

namespace networking
{
namespace message
{

using SendHandler = std::function<void(const error::ErrorCode & code)>;

template<typename Message>
using ReceiveHandler = std::function<void(const error::ErrorCode & code, Message & message)>;

using SendToHandler = std::function<void(const error::ErrorCode & code)>;

template<typename Message>
using ReceiveFromHandler = std::function<
    void(const error::ErrorCode & code,
         Message & message,
         const std::string & senderHost,
         std::uint16_t senderPort)>;

template<typename Message>
struct Encoder;

template<>
struct Encoder<std::string>
{
    std::string operator()(const std::string & message) const
    { return message; }
};

template<typename Message>
struct Decoder;

template<>
struct Decoder<std::string>
{
    std::string operator()(const std::string & data) const
    { return data; }
};

namespace internal
{

// Some pre-processing functions before sending the stream over the network.
// Since our message delimiter is '\0' the message itself must not contain such a character.
// Therefore we will replace all occurrences of '\0' with "\0" ("\" is the escape character).
// This implies that we must also change "\" to "\\".

template<typename Message>
std::string encode(const Message & message)
{
    auto data = Encoder<Message>{}(message);
    boost::replace_all(data, "\\", "\\\\");
    boost::replace_all(data, std::string{'\0'}, "\\0");
    return data;
}

template<typename Message>
Message decode(std::string & data)
{
    boost::replace_all(data, "\\\\", "\\");
    boost::replace_all(data, "\\0", std::string{'\0'});
    return Decoder<Message>{}(data);
}

}

template<typename Message>
void send(Networking & net,
          boost::asio::ip::tcp::socket & socket,
          const Message & message,
          const time::Duration & timeout)
{
    using namespace std::chrono_literals;
    if (timeout <= 0s)
        throw error::Aborted();

    networking::stream::write(net, socket, internal::encode(message), timeout);
};

template<typename Message>
void asyncSend(Networking & net,
               boost::asio::ip::tcp::socket & socket,
               const Message & message,
               const time::Duration & timeout,
               SendHandler handler)
{
    auto data = std::make_shared<std::string>(internal::encode(message));
    networking::stream::asyncWrite(
        net, socket, *data, timeout,
        [handler, data](const auto & errorCode)
        { handler(errorCode); });
};

template<typename Message>
Message receive(Networking & net,
                boost::asio::ip::tcp::socket & socket,
                boost::asio::streambuf & buffer,
                const time::Duration & timeout)
{
    using namespace std::chrono_literals;
    if (timeout <= 0s)
        throw error::Aborted();

    auto data = networking::stream::read(net, socket, buffer, timeout);
    return internal::decode<Message>(data);
};

template<typename Message>
void asyncReceive(Networking & net,
                  boost::asio::ip::tcp::socket & socket,
                  boost::asio::streambuf & buffer,
                  const time::Duration & timeout,
                  ReceiveHandler<Message> handler)
{
    networking::stream::asyncRead(
        net, socket, buffer, timeout,
        [handler](const auto & errorCode, auto & data)
        {
            auto message = internal::decode<Message>(data);
            handler(errorCode, message);
        });
};

template<typename Message>
void sendDatagramTo(Networking & net,
                    boost::asio::ip::udp::socket & socket,
                    const Message & message,
                    const std::string & host,
                    std::uint16_t port,
                    const time::Duration & timeout)
{
    networking::socket::sendTo(net, socket, internal::encode(message), host, port, timeout);
}

template<typename Message>
void asyncSendDatagramTo(Networking & net,
                         boost::asio::ip::udp::socket & socket,
                         const Message & message,
                         const std::string & host,
                         std::uint16_t port,
                         const time::Duration & timeout,
                         SendToHandler handler)
{
    auto data = std::make_shared<std::string>(internal::encode(message));
    networking::socket::asyncSendTo(
        net, socket, *data, host, port, timeout,
        [handler, data](const auto & error)
        { handler(error); });
}

template<typename Message>
Message receiveDatagramFrom(Networking & net,
                            boost::asio::ip::udp::socket & socket,
                            std::vector<char> & buffer,
                            std::string & host,
                            std::uint16_t & port,
                            const time::Duration & timeout)
{
    auto data = networking::socket::receiveFrom(net, socket, buffer, host, port, timeout);
    return internal::decode<Message>(data);
}

template<typename Message>
void asyncReceiveDatagramFrom(Networking & net,
                              boost::asio::ip::udp::socket & socket,
                              std::vector<char> & buffer,
                              const time::Duration & timeout,
                              ReceiveFromHandler<Message> handler)
{
    networking::socket::asyncReceiveFrom(
        net, socket, buffer, timeout,
        [handler](auto error, auto & data, const auto & senderHost, auto senderPort)
        {
            auto message = internal::decode<Message>(data);
            handler(error, message, senderHost, senderPort);
        });
}

}
}


#endif //PROTOCOL_MESSAGES_H
