//
// Created by philipp on 02.01.18.
//

#ifndef NETWORKINGLIB_SOCKETOPS_H
#define NETWORKINGLIB_SOCKETOPS_H

#include "Stream.h"
#include "Resolver.h"

namespace networking
{
namespace socket
{

namespace internal
{

boost::asio::const_buffers_1 stringToBuffer(const std::string & str);

bool isLastCharacterNull(const std::vector<char> & buffer, std::size_t numBytes);

std::string stringFromBuffer(const std::vector<char> & buffer, std::size_t numBytes);

}

using ConnectHandler = std::function<void(const error::ErrorCode & error)>;

using SendHandler = std::function<void(const error::ErrorCode & error)>;

using ReceiveHandler = std::function<void(const error::ErrorCode & error,
                                          std::string & data,
                                          const std::string & host,
                                          std::uint16_t port)>;

template<typename SocketService>
void connect(Networking & net,
             SocketService & socket,
             const std::string & host,
             std::uint16_t port,
             time::Duration timeout)
{
    using Resolver = networking::internal::CloseableResolver<boost::asio::ip::tcp>;

    auto startTime = time::now();

    // Resolve host.
    Resolver resolver{net.getIoService()};
    Resolver::Query query{host, std::to_string(port)};

    auto resolveOperation = [&resolver](auto && ... args)
    { resolver.async_resolve(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        Resolver::Iterator> resolveResult;

    closeable::timedOperation(
        resolveResult,
        net,
        resolveOperation,
        resolver,
        timeout,
        query);

    auto endpointIterator = std::get<1>(resolveResult);

    // Update timeout.
    auto timeSpend = time::now() - startTime;
    timeout -= timeSpend;

    auto connectOperation = [](auto && ... args)
    { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        Resolver::Iterator> connectResult;

    closeable::timedOperation(
        connectResult,
        net,
        connectOperation,
        socket,
        timeout,
        socket,
        endpointIterator);
}

template<typename SocketService>
void asyncConnect(Networking & net,
                  SocketService & socket,
                  const std::string & host,
                  std::uint16_t port,
                  const time::Duration & timeout,
                  const ConnectHandler & handler)
{
    using Resolver = networking::internal::CloseableResolver<boost::asio::ip::tcp>;

    auto startTime = time::now();

    // Resolve host.
    auto resolver = std::make_shared<Resolver>(net.getIoService());
    Resolver::Query query{host, std::to_string(port)};

    auto resolveOperation = [&resolver](auto && ... args)
    { resolver->async_resolve(std::forward<decltype(args)>(args)...); };

    closeable::timedAsyncOperation(
        net,
        resolveOperation,
        resolver,
        timeout,
        [&net, &socket, host, port, timeout, handler, resolver, startTime]
            (const auto & networkingError, const auto & boostError, auto endpointIterator)
        {
            if (networkingError)
            {
                handler(networkingError);
                return;
            }

            // Update timeout.
            auto timeSpend = time::now() - startTime;
            auto newTimeout = timeout - timeSpend;

            auto connectOperation = [](auto && ... args)
            { boost::asio::async_connect(std::forward<decltype(args)>(args)...); };

            closeable::timedAsyncOperation(
                net,
                connectOperation,
                socket,
                newTimeout,
                [handler](const auto & networkingError, const auto & boostError, auto iterator)
                { handler(networkingError); },
                socket,
                endpointIterator);
        },
        query);
}

template<typename DatagramSocket>
void sendTo(Networking & net,
            DatagramSocket & socket,
            const std::string & data,
            const std::string & host,
            std::uint16_t port,
            const time::Duration & timeout)
{
    boost::asio::ip::udp::endpoint endpoint{boost::asio::ip::address::from_string(host), port};

    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_send_to(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        std::size_t> result;

    closeable::timedOperation(
        result,
        net,
        asyncOperation,
        socket,
        timeout,
        internal::stringToBuffer(data),
        endpoint);

    auto bytesTransferred = std::get<1>(result);
    if (bytesTransferred < data.size())
        throw error::FailedOperation{};
};

template<typename DatagramSocket>
void asyncSendTo(Networking & net,
                 DatagramSocket & socket,
                 const std::string & data,
                 const std::string & host,
                 std::uint16_t port,
                 const time::Duration & timeout,
                 const SendHandler & handler)
{
    boost::asio::ip::udp::endpoint endpoint{boost::asio::ip::address::from_string(host), port};

    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_send_to(std::forward<decltype(args)>(args)...); };

    closeable::timedAsyncOperation(
        net,
        asyncOperation,
        socket,
        timeout,
        [&data, handler](const auto & networkingError, const auto & boostError, auto bytesTransferred)
        {
            if (bytesTransferred < data.size())
            {
                handler(error::codes::FAILED_OPERATION);
                return;
            }

            handler(networkingError);
        },
        internal::stringToBuffer(data),
        endpoint);
};

template<typename DatagramSocket>
std::string receiveFrom(Networking & net,
                        DatagramSocket & socket,
                        std::vector<char> buffer,
                        std::string & senderHost,
                        std::uint16_t & senderPort,
                        const time::Duration & timeout)
{
    boost::asio::ip::udp::endpoint senderEndpoint;

    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_receive_from(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        std::size_t> result;

    closeable::timedOperation(
        result,
        net,
        asyncOperation,
        socket,
        timeout,
        boost::asio::buffer(buffer),
        senderEndpoint);

    senderHost = senderEndpoint.address().to_string();
    senderPort = senderEndpoint.port();

    auto bytesTransferred = std::get<1>(result);
    if (!internal::isLastCharacterNull(buffer, bytesTransferred))
        throw error::FailedOperation{};

    return internal::stringFromBuffer(buffer, bytesTransferred);
}

template<typename DatagramSocket>
void asyncReceiveFrom(Networking & net,
                      DatagramSocket & socket,
                      std::vector<char> & buffer,
                      const time::Duration & timeout,
                      ReceiveHandler handler)
{
    auto senderEndpoint = std::make_shared<boost::asio::ip::udp::endpoint>();

    auto asyncOperation = [&socket](auto && ... args)
    { socket.async_receive_from(std::forward<decltype(args)>(args)...); };

    closeable::timedAsyncOperation(
        net,
        asyncOperation,
        socket,
        timeout,
        [&buffer, handler, senderEndpoint](const auto & networkingError, const auto & boostError, auto bytesTransferred)
        {
            if (!internal::isLastCharacterNull(buffer, bytesTransferred))
            {
                std::string data{};
                std::string senderHost{};
                std::uint16_t senderPort{};
                handler(error::codes::FAILED_OPERATION, data, senderHost, senderPort);
                return;
            }

            auto senderHost = senderEndpoint->address().to_string();
            auto senderPort = senderEndpoint->port();
            auto data = internal::stringFromBuffer(buffer, bytesTransferred);
            handler(networkingError, data, senderHost, senderPort);
        },
        boost::asio::buffer(buffer),
        *senderEndpoint);
}

}
}

#endif //NETWORKINGLIB_SOCKETOPS_H
