//
// Created by philipp on 02.12.17.
//

#ifndef PROTOCOL_NETWORKUTILS_H
#define PROTOCOL_NETWORKUTILS_H

#include <boost/asio/ip/tcp.hpp>
#include <boost/asio.hpp>
#include "Networking.h"
#include "Timer.h"
#include "Error.h"
#include "Closeable.h"

namespace networking
{
namespace stream
{

namespace internal
{

boost::asio::const_buffers_1 stringToBuffer(const std::string & str);

std::string stringFromStreambuf(boost::asio::streambuf & streambuf, std::size_t numBytes);

}

using WriteHandler = std::function<void(const error::ErrorCode & error)>;

using ReadHandler = std::function<void(const error::ErrorCode & error, std::string & data)>;

template<typename SyncWriteStream>
void write(Networking & net,
           SyncWriteStream & stream,
           const std::string & data,
           const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        std::size_t> result;

    closeable::timedOperation(
        result,
        net,
        asyncOperation,
        stream,
        timeout,
        stream,
        internal::stringToBuffer(data));

    auto bytesTransferred = std::get<1>(result);
    if (bytesTransferred < data.size())
        throw error::FailedOperation{};
};

template<typename SyncWriteStream>
void asyncWrite(Networking & net,
                SyncWriteStream & stream,
                const std::string & data,
                const time::Duration & timeout,
                const WriteHandler & handler)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_write(std::forward<decltype(args)>(args)...); };

    closeable::timedAsyncOperation(
        net,
        asyncOperation,
        stream,
        timeout,
        [&data, handler](const auto & networkingError, const auto & boostError, auto bytesTransferred)
        {
            if (bytesTransferred < data.size())
            {
                handler(error::codes::FAILED_OPERATION);
                return;
            }

            handler(networkingError);
        },
        stream,
        internal::stringToBuffer(data));
}

template<typename SyncReadStream>
std::string read(Networking & net,
          SyncReadStream & stream,
          boost::asio::streambuf & buffer,
          const time::Duration & timeout)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read_until(std::forward<decltype(args)>(args)...); };

    std::tuple<
        boost::system::error_code,
        std::size_t> result;

    closeable::timedOperation(
        result,
        net,
        asyncOperation,
        stream,
        timeout,
        stream,
        buffer,
        '\0');

    auto bytesTransferred = std::get<1>(result);
    return internal::stringFromStreambuf(buffer, bytesTransferred);
};

template<typename SyncReadStream>
void asyncRead(Networking & net,
               SyncReadStream & stream,
               boost::asio::streambuf & buffer,
               const time::Duration & timeout,
               const ReadHandler & handler)
{
    auto asyncOperation = [](auto && ... args)
    { boost::asio::async_read_until(std::forward<decltype(args)>(args)...); };

    closeable::timedAsyncOperation(
        net,
        asyncOperation,
        stream,
        timeout,
        [&buffer, handler](const auto & networkingError, const auto & boostError, auto bytesTransferred)
        {
            auto data = internal::stringFromStreambuf(buffer, bytesTransferred);
            handler(networkingError, data);
        },
        stream,
        buffer,
        '\0');
};

}
}

#endif //PROTOCOL_NETWORKUTILS_H
