//
// Created by philipp on 02.12.17.
//

#include "Test.h"
#include "TestUtils.h"
#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/ServiceServer.h"
#include "PlatoonService.h"
#include "../include/ServiceClient.h"
#include "../include/DatagramReceiver.h"
#include "../include/DatagramSender.h"

using boost::asio::ip::tcp;

using namespace std::chrono_literals;

namespace networking
{
namespace test
{

void testServices()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto server = service::Server<PlatoonService>::create(net1, 10001);

    server->advertiseService(
        [](const auto & clientEndpoint,
           auto & requestMessage) -> PlatoonMessage
        {
            std::cout << "Request from "
                      << (int) requestMessage.getVehicleId()
                      << " with type: "
                      << (int) requestMessage.getMessageType()
                      << "\n";
            return PlatoonMessage::acceptResponse(1, 42);
        });

    sleep(1);

    auto client = service::Client<PlatoonService>::create(net2);
    for (int i = 0; i < 5; i++)
    {
        auto response = client->call("127.0.0.1", 10001, PlatoonMessage::followerRequest(2), 5s);
        std::cout << "Response from " << (int) response.getVehicleId() << " with type: "
                  << (int) response.getMessageType()
                  << " and platoonId: " << (int) response.getPlatoonId() << "\n";
    }
}

void testTcpClientTimeout()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    const auto timeout = 3s;

    auto server = service::Server<PlatoonService>::create(net1, 10001);

    server->advertiseService(
        [](const auto & clientEndpoint,
           auto & requestMessage) -> PlatoonMessage
        {
            // Just sleep for 5 seconds.
            sleep(5);
            return PlatoonMessage::acceptResponse(1, 42);
        });

    sleep(1);

    auto client = service::Client<PlatoonService>::create(net2);
    auto startTime = boost::posix_time::microsec_clock::local_time();
    try
    {
        auto response = client->call("127.0.0.1", 10001, PlatoonMessage::followerRequest(2), timeout);
        std::cout << "Response: " << response.getPlatoonId() << "\n";
        std::cout << "FAILED!";
    }
    catch (const error::Aborted & e)
    {
        auto nowTime = boost::posix_time::microsec_clock::local_time();
        auto timeSpend = nowTime - startTime;
        if (timeSpend.seconds() >= 2)
            std::cout << "SUCCESS!\n";
    }
}

void testMultipleConnections()
{
    using namespace protocol;
    Networking net1;
    Networking net2;
    Networking net3;

    auto server1 = service::Server<PlatoonService>::create(net1, 10001);
    auto server2 = service::Server<PlatoonService>::create(net2, 10002);

    server1->advertiseService(
        [](const auto & clientEndpoint,
           auto & requestMessage) -> PlatoonMessage
        {
            return PlatoonMessage::acceptResponse(1, 42);
        });
    server2->advertiseService(
        [](const auto & clientEndpoint,
           auto & requestMessage) -> PlatoonMessage
        {
            return PlatoonMessage::acceptResponse(2, 43);
        });

    sleep(1);

    auto client = service::Client<PlatoonService>::create(net3);

    auto response = client->call("127.0.0.1", 10001, PlatoonMessage::followerRequest(1), 5s);
    std::cout << "Response from " << response.getVehicleId() << std::endl;

    response = client->call("127.0.0.1", 10002, PlatoonMessage::followerRequest(2), 5s);
    std::cout << "Response from " << response.getVehicleId() << std::endl;
}

void testStoppingServiceServer()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto server = service::Server<PlatoonService>::create(net1, 10001);

    auto handler = [](const auto & clientEndpoint,
                      auto & requestMessage) -> PlatoonMessage
    {
        // Just sleep for 3 seconds.
        sleep(2);
        return PlatoonMessage::acceptResponse(1, 42);
    };

    server->advertiseService(handler);

    sleep(1);

    auto client = service::Client<PlatoonService>::create(net2);
    try
    {
        client->call("127.0.0.1", 10001, PlatoonMessage::followerRequest(42), 1s);
    }
    catch (const error::Aborted & e)
    {
        server->stop();
        server->advertiseService(handler);
        auto response = client->call("127.0.0.1", 10001, PlatoonMessage::followerRequest(42), 5s);
        if (response.getMessageType() == messageTypes::ACCEPT_RESPONSE)
            std::cout << "SUCCESS!\n";
    }
}

void testAsyncDatagramReceiver()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto receiver = message::DatagramReceiver<PlatoonMessage>::create(net1, 10000);
    auto sender = message::DatagramSender<PlatoonMessage>::create(net2);

    std::atomic<bool> running{true};

    receiver->asyncReceive(
        3s,
        [&running](const auto & error,
                   auto & message,
                   const auto & senderHost,
                   auto senderPort)
        {
            if (!error)
                std::cout << "SUCCESS! Received message from: " << message.getVehicleId() << "\n";

            running = false;
        });

    sleep(1);

    sender->send(PlatoonMessage::followerRequest(42), "127.0.0.1", 10000, 5s);

    while (running);
}

void testPeriodicTimer()
{
    Networking net;

    auto timer = time::Timer::create(net);

    int run = 0;
    std::atomic<bool> running{true};
    auto startTime = time::now();
    timer->startPeriodicTimeout(
        1s,
        [&run, &running, timer, &startTime]()
        {
            if (run >= 3)
            {
                std::cout << "SUCCESS!\n";
                timer->stop();
                running = false;
                return;
            }

            auto deltaTime = std::chrono::duration_cast<std::chrono::milliseconds>(time::now() - startTime);
            startTime = time::now();
            std::cout << "Delta time [ms]: " << deltaTime.count() << "\n";
            if (std::abs(1s - deltaTime) > 2ms)
                std::cout << "FAILED!\n";
            run++;
        });

    while (running);
}

void testServiceClientAsyncCallTimeout()
{
    using namespace protocol;
    Networking net1;
    Networking net2;

    auto server = service::Server<PlatoonService>::create(net1, 10001);

    server->advertiseService(
        [](const auto & clientEndpoint,
           auto & requestMessage) -> PlatoonMessage
        {
            sleep(3);
            return PlatoonMessage::acceptResponse(1, 42);
        });

    sleep(1);

    auto client = service::Client<PlatoonService>::create(net2);

    std::atomic<bool> running{true};

    client->asyncCall(
        "127.0.0.1", 10001, PlatoonMessage::followerRequest(1), 1s,
        [&running](const auto & error, const auto & response)
        {
            if (error == error::codes::ABORTED)
                std::cout << "SUCCESS!\n";

            running = false;
        });

    while (running);
}

void testDatagramSenderAsyncSend()
{
    using namespace protocol;
    Networking net;

    auto receiver = message::DatagramReceiver<PlatoonMessage>::create(net, 10000);
    auto sender = message::DatagramSender<PlatoonMessage>::create(net);

    std::atomic<bool> running{true};

    receiver->asyncReceive(
        3s,
        [&running](const auto & error,
                   auto & message,
                   const std::string & senderHost,
                   auto senderPort)
        {
            if (error)
                std::cout << "FAILED! (receive error)\n";

            std::cout << "Sender host: " << senderHost << "\nSender Port: " << senderPort << "\n";
            if (message.getPlatoonId() == 42)
                std::cout << "SUCCESS!\n";

            running = false;
        });

    sender->asyncSend(
        PlatoonMessage::acceptResponse(1, 42), "127.0.0.1", 10000, 1s,
        [](const auto & error)
        {
            if (error)
                std::cout << "FAILED! (send error)\n";
        });

    while (running);
}

void testResolver()
{
    Networking net;

    auto resolver = Resolver::create(net);

    std::atomic<bool> running{true};
    bool thrown{false};

    auto endpoints = resolver->resolve("google.de", "http", 5s);
    for (const auto & endpoint : endpoints)
        std::cout << "ip: " << endpoint.ip << " port: " << endpoint.port << "\n";

    resolver->asyncResolve(
        "google.de", "http", 5s,
        [&running](const auto & error, const auto & endpoints)
        {
            for (const auto & endpoint : endpoints)
                std::cout << "ip: " << endpoint.ip << " port: " << endpoint.port << "\n";

            running = false;
        });

    try
    {
        resolver->resolve("google.de", "http", 5s);
    }
    catch (const error::Busy & error)
    {
        thrown = true;
    }

    while (running);

    if (thrown)
        std::cout << "SUCCESS!\n";
}

}
}

